adjectives = [
    "able", "abnormal", "admiring", "adoring", "adventurous",
    "affectionate", "agile", "agitated", "agreeable", "alert",
    "amazing", "ambitious", "amiable", "amusing", "analytical",
    "angelic", "angry", "apathetic", "apprehensive", "ardent",
    "artificial", "artistic", "assertive", "attentive", "average",
    "awesome", "awful", "balanced", "beautiful", "beneficent",
    "blissful", "blue", "blunt", "boisterous", "boring", "brave",
    "bright", "brilliant", "buff", "callous", "candid",
    "cantankerous", "capable", "careful", "careless", "caustic",
    "cautious", "charming", "cheerful", "chic", "childish",
    "childlike", "churlish", "circumspect", "civil", "clean",
    "clever", "clumsy", "cocky", "coherent", "cold", "compassionate",
    "competent", "composed", "conceited", "condescending",
    "confident", "confused", "conscientious", "considerate",
    "content", "cool", "cooperative", "cordial", "courageous",
    "cowardly", "crabby", "crafty", "cranky", "crass", "critical",
    "cruel", "curious", "cynical", "dainty", "dazzling", "decisive",
    "deep", "deferential", "deft", "delicate", "delightful",
    "demonic", "demure", "dependent", "depressed", "determined",
    "devoted", "dextrous", "diligent", "direct", "dirty",
    "disagreeable", "discerning", "discreet", "disruptive", "distant",
    "distracted", "distraught", "distrustful", "dowdy", "dramatic",
    "dreamy", "dreary", "drowsy", "drugged", "drunk", "dull",
    "dutiful", "eager", "earnest", "ecstatic", "efficient",
    "egotistical", "elastic", "elated", "elegant", "elfin",
    "eloquent", "emotional", "energetic", "enterprising",
    "enthusiastic", "epic", "evasive", "exacting", "excellent",
    "excitable", "experienced", "fabulous", "fastidious", "ferocious",
    "fervent", "festive", "fiery", "flabby", "flaky", "flamboyant",
    "flashy", "focused", "frank", "friendly", "frosty", "funny",
    "fussy", "gallant", "generous", "gentle", "gifted", "gloomy",
    "glutinous", "good", "goofy", "gracious", "grave", "great",
    "groggy", "grouchy", "guarded", "happy", "hardcore", "hateful",
    "hearty", "helpful", "hesitant", "heuristic", "hopeful", "hungry",
    "hypercritical", "hysterical", "idiotic", "idle", "illogical",
    "imaginative", "immature", "immodest", "impatient",
    "imperturbable", "impetuous", "impractical", "impressionable",
    "impressive", "impulsive", "inactive", "incisive", "incompetent",
    "inconsiderate", "inconsistent", "indefatigable", "independent",
    "indiscreet", "indolent", "industrious", "inexperienced",
    "infallible", "insensitive", "inspiring", "intelligent",
    "interesting", "intolerant", "inventive", "irascible",
    "irritable", "irritating", "jocular", "jolly", "jovial", "joyous",
    "judgmental", "keen", "kind", "lame", "laughing", "lazy", "lean",
    "leery", "lethargic", "listless", "lithe", "lively", "local",
    "logical", "lovable", "lovely", "loving", "lucid", "maternal",
    "mature", "mean", "meddlesome", "mercurial", "methodical",
    "meticulous", "mild", "miserable", "modest", "moronic", "morose",
    "motivated", "musical", "musing", "mystifying", "naive", "nasty",
    "natural", "naughty", "negative", "nervous", "nifty", "noisy",
    "normal", "nostalgic", "nosy", "numb", "objective", "obliging",
    "obnoxious", "optimistic", "orderly", "ostentatious", "outgoing",
    "outspoken", "passionate", "passive", "paternal", "paternalistic",
    "patient", "peaceful", "pedantic", "peevish", "pensive",
    "persevering", "persnickety", "petulant", "picky", "plain",
    "playful", "pleasant", "plucky", "polite", "popular", "positive",
    "powerful", "practical", "prejudiced", "pretty", "priceless",
    "proficient", "proud", "provocative", "prudent", "punctual",
    "quarrelsome", "querulous", "quick", "quiet", "quirky",
    "quizzical", "realistic", "reassuring", "reclusive", "relaxed",
    "reliable", "reluctant", "resentful", "reserved", "resigned",
    "resourceful", "respected", "respectful", "responsible",
    "restless", "revered", "reverent", "ridiculous", "romantic",
    "sad", "sassy", "saucy", "sedate", "selfish", "sensible",
    "sensitive", "sentimental", "serene", "serious", "sharp",
    "shrewd", "shy", "silly", "sincere", "sleepy", "slight", "sloppy",
    "slothful", "slovenly", "slow", "smart", "snazzy", "sneering",
    "snobby", "sober", "somber", "sophisticated", "soulful",
    "soulless", "sour", "spirited", "spiteful", "stable", "staid",
    "steady", "stern", "stoic", "striking", "strong", "stupefied",
    "stupid", "sturdy", "subtle", "sulky", "sullen", "supercilious",
    "superficial", "surly", "suspicious", "sweet", "tactful",
    "tactless", "talented", "tender", "testy", "thinking", "thirsty",
    "thoughtful", "thoughtless", "timid", "tired", "tolerant",
    "touchy", "tranquil", "trusting", "ugly", "unaffected",
    "unbalanced", "uncertain", "uncooperative", "undependable",
    "unemotional", "unfriendly", "unguarded", "unhelpful",
    "unimaginative", "unmotivated", "unpleasant", "unpopular",
    "unreliable", "unruffled", "unsophisticated", "unsure",
    "unthinking", "unwilling", "upbeat", "venal", "versatile",
    "vibrant", "vigilant", "vigorous", "volcanic", "vulnerable",
    "warm", "warmhearted", "wary", "watchful", "weak", "willing",
    "wizardly", "wonderful", "youthful", "zealous"
]

creatures = [
    "accentor", "acouchi", "addax", "agama", "agouti", "albatross",
    "alligator", "alpaca", "anaconda", "angelfish", "anhinga", "anoa",
    "anole", "ant", "anteater", "antelope", "archerfish", "armadillo",
    "arowana", "auk", "avocet", "axolotl", "babirusa", "baboon",
    "badger", "bandicoot", "banshee", "banteng", "barbet", "basilisk",
    "bat", "batfish", "beaver", "bee", "beetle", "betta", "bettong",
    "bigeyes", "bilby", "binturong", "bird", "bison", "bitterling",
    "bittern", "blackdevil", "bluebird", "boa", "bongo", "booby",
    "bowerbird", "broadbill", "bulbul", "bunting", "bushbaby",
    "bushbaby", "bushshrike", "bustard", "butterfly", "buzzard",
    "caecilian", "camel", "capybara", "caracal", "caracara",
    "cardinal", "cardinalfish", "caribou", "caribou", "carp",
    "cassowary", "cat", "caterpillar", "catfish", "cattle", "cattle",
    "cattle", "cavy", "centaur", "centipede", "cerberus", "chameleon",
    "chamois", "charybdis", "chat", "cheetah", "chicken", "chimaera",
    "chimera", "chimpanzee", "chinchilla", "chipmunk", "chough",
    "chuckwalla", "cicada", "cichlid", "civet", "clam", "clownfish",
    "coati", "cobra", "cockatiel", "cockatoo", "cockroach", "colugo",
    "conch", "coot", "copperhead", "cormorant", "cotinga", "cougar",
    "cougar", "courser", "coyote", "coypu", "coypu", "crab", "crane",
    "crayfish", "cricket", "crocodile", "crow", "cuckoo", "culpeo",
    "curlew", "cuscus", "cuttlefish", "cynocephalus", "dartfish",
    "deer", "degu", "demon", "dhole", "dibbler", "dikkop", "dikkop",
    "dikkop", "dimetrodon", "dingo", "dinosaur", "dipper", "discus",
    "dodo", "dog", "dolphin", "donkey", "doppelganger", "dormouse",
    "dove", "dragonfly", "drongo", "duck", "dugong", "duiker",
    "dunnart", "eagle", "echidna", "echidna", "eel", "egret", "eider",
    "elephant", "elk", "elk", "emu", "fairy", "falcon", "ferret",
    "finch", "fish", "fisher", "flamingo", "flea", "flowerpecker",
    "fly", "fossa", "fox", "frigatebird", "frog", "frogmouth",
    "fulmar", "gannet", "gar", "gaur", "gazelle", "gecko", "gerbil",
    "gibbon", "giraffe", "goat", "goatfish", "goblin", "goldfish",
    "golem", "goose", "gopher", "gopher", "goral", "gorilla",
    "gourami", "grackle", "grasshopper", "grebe", "griffin", "grison",
    "groundhog", "groundhog", "grouse", "guanaco", "gull", "gull",
    "gundi", "hamster", "harrier", "hartebeest", "hawk", "hawkodile",
    "hawkodile", "hedgehog", "helmetshrike", "heron", "hippopotamus",
    "honeyeater", "hornbill", "hornet", "horse", "hoverfly",
    "hummingbird", "hutia", "hyena", "hyrax", "ibex", "ibis",
    "icterid", "iguana", "impala", "insect", "jacana", "jack",
    "jackal", "jaguar", "jaguarundi", "jay", "jellyfish", "jerboa",
    "kangaroo", "kerodon", "kestrel", "kingbird", "kingfisher",
    "kinkajou", "kite", "kitten", "kiwi", "klipspringer", "knifefish",
    "koala", "kodkod", "kookaburra", "kowari", "kudu", "kultarr",
    "ladybug", "lamprey", "lapwing", "leech", "lemming", "lemur",
    "leopard", "leprechaun", "liger", "lion", "lionfish", "lizard",
    "llama", "loach", "lobster", "longspur", "loon", "loris", "lory",
    "lovebird", "lynx", "lyrebird", "macaw", "mamba", "mammoth",
    "manakin", "manatee", "mandrill", "manticore", "mara", "margay",
    "markhor", "marlin", "marmot", "marten", "mastodon", "meadowlark",
    "meerkat", "megaloceros", "megapode", "mermaids", "millipede",
    "mink", "minnow", "minotaur", "mockingbird", "mole", "mongoose",
    "monkey", "moorhen", "moorhen", "moose", "mosasaur", "mosquito",
    "moth", "mothman", "motmot", "mouse", "mousebird", "mudpuppy",
    "mudskipper", "mullet", "muntjac", "muskox", "muskrat", "mutants",
    "narwhal", "nautilus", "needlefish", "newt", "nighthawk",
    "nightjar", "numbat", "nuthatch", "ocelot", "octopus", "okapi",
    "olingo", "onager", "opossum", "orangutan", "orca", "oriole",
    "orthros", "oryx", "osprey", "ostrich", "otter", "ovenbird",
    "owl", "oyster", "paca", "paddlefish", "pademelon", "pangolin",
    "panther", "parakeet", "parakeet", "parrot", "parrotfish",
    "peafowl", "peafowl", "peccary", "pelican", "penguin", "pheasant",
    "pig", "pigeon", "pika", "pike", "piranha", "pitohui", "platypus",
    "plover", "pony", "porcupine", "porpoise", "possum", "potoo",
    "potoroo", "potto", "pronghorn", "ptarmigan", "pterosaur", "pudu",
    "puffin", "puma", "puppy", "python", "quagga", "quail", "quetzal",
    "quokka", "quoll", "rabbit", "rabbit", "raccoon", "rail",
    "rainbowfish", "rat", "rattlesnake", "raven", "rhea",
    "rhinoceros", "ringtail", "roadrunner", "robin", "rockfish",
    "roller", "rook", "sailfish", "salamander", "salmon",
    "sandgrouse", "sawfish", "scorpion", "scylla", "seadragon",
    "seahorse", "seal", "serval", "shapeshifter", "shark",
    "shearwater", "sheep", "sheep", "shelduck", "shoebill", "shrew",
    "shrimp", "shrimp", "siren", "skate", "skimmer", "skink", "skua",
    "skunk", "sloth", "slug", "snail", "snake", "snipe", "solenodon",
    "songbird", "sparrow", "spider", "sponge", "spoonbill",
    "springhare", "squid", "squirrel", "starfish", "starling",
    "stingray", "stonefish", "stork", "sturgeon", "sunbird",
    "surgeonfish", "swallow", "swamphen", "swan", "swift",
    "swordfish", "tadpole", "takin", "tamandua", "tamarin", "tanager",
    "tapaculo", "tapir", "tarantula", "tarpon", "tarsier", "tayra",
    "tenrec", "termite", "tern", "tetra", "thornbill", "thrasher",
    "thrush", "thunderbird", "tiger", "toad", "toadfish", "tortoise",
    "toucan", "treeshrew", "trogon", "trout", "trumpeter", "tuatara",
    "tuna", "turaco", "turkey", "turtle", "typhon", "unicorn",
    "urchin", "vampire", "vanga", "vaquita", "vicuna", "viper",
    "viscacha", "vole", "vulture", "wader", "wagtail", "wallaby",
    "walleye", "walrus", "warbler", "warthog", "wasp", "waxwing",
    "weasel", "weaver", "wendigo", "werewolf", "whale", "whistler",
    "whydah", "wigeon", "wildcat", "wildebeest", "wolf",
    "wolverine", "wombat", "woodcock", "woodpecker", "woodswallow",
    "worm", "wren", "xenopoecilus", "xenops", "xerus", "yak",
    "yellowjacket", "zebra", "zebu", "zombie"
]
